! ===============================================================================================
MODULE ACTPol_deep 
! ===============================================================================================
  use Options
  use Fg_subroutines
  use Foregrounds_loading

  implicit none
  logical :: initialise_actpol_deep=.true.

  character(len=*), parameter, public :: actpol_likelihood_version='ACTPol_deep_like' 

  real(8), dimension(:), allocatable :: b_dat
  real(8), dimension(:,:), allocatable :: fisher, covmat
  real(8), dimension(:,:), allocatable :: win_func_tt, win_func_te, win_func_ee
  real(8), dimension(:), allocatable :: diff_vec
  real(8), dimension(:), allocatable :: l98, l150, ell

  PRIVATE
  public :: actpol_deep_likelihood_init
  public :: actpol_deep_likelihood_compute
  public :: get_fg

contains

! ===============================================================================================
  SUBROUTINE actpol_deep_likelihood_init(spec_fname, cov_fname, bbltt_fname, bblte_fname, bblee_fname, data_dir, act_data_dir, nbint, lmax_win, bmax0, b1tt, b1te, b1ee, data_run, leakage_correction_fname)
! -----------------------------------------------------------------------------------------------
    
  implicit none
  integer :: i, j, lun, il, info
  character(len=240) :: spec_filename, cov_filename
  character(len=240) :: bbltt_filename, bblte_filename, bblee_filename
  character(len=240) :: spec_fname, cov_fname
  character(len=240) :: bbltt_fname, bblte_fname, bblee_fname
  character(len=240) :: data_dir, act_data_dir
  character(len=240) :: leakage_correction_fname, leakage_correction_filename
  integer :: nbint, bmax0, b1tt, b1te, b1ee
  integer :: tt_lmax, lmax_win
  integer :: nbins_leak ! # bins used for leakage correction = # bins for TE, EE
  logical :: data_run
    
  nbins_leak = bmax0 - b1te + 1

  print *, 'Initializing ACTPol likelihood, version '//actpol_likelihood_version
  !-----------------------------------------------
  ! set file names
  !-----------------------------------------------
   spec_filename  = trim(data_dir)//trim(ACT_data_dir)//trim(spec_fname)
   cov_filename   = trim(data_dir)//trim(ACT_data_dir)//trim(cov_fname)
   bbltt_filename = trim(data_dir)//trim(ACT_data_dir)//trim(bbltt_fname)
   bblte_filename = trim(data_dir)//trim(ACT_data_dir)//trim(bblte_fname)
   bblee_filename = trim(data_dir)//trim(ACT_data_dir)//trim(bblee_fname)
   leakage_correction_filename = trim(data_dir)//trim(ACT_data_dir)//trim(leakage_correction_fname)

  !-----------------------------------------------
  ! load spectrum
  !-----------------------------------------------
   allocate(b_dat(1:nbint)) 
   call get_free_lun(lun)
   open(unit=lun,file=spec_filename,form='formatted',status='unknown',action='read')
   do i=1,nbint
      read(lun,*) b_dat(i)
   enddo
   close(lun) 

  !-----------------------------------------------
  ! load binning matrices
  !-----------------------------------------------
   ! TT
   call get_free_lun(lun)
   open(unit=lun,file=bbltt_filename,form='formatted',status='unknown',action='read')
   allocate(win_func_tt(1:bmax0,1:lmax_win)) !Defined over ACTPol's full ell range
   do i=1,bmax0
      read(lun,*) (win_func_tt(i,il), il=2,lmax_win)
   enddo
   close(lun)

   ! TE
   call get_free_lun(lun)
   open(unit=lun,file=bblte_filename,form='formatted',status='unknown',action='read')
   allocate(win_func_te(1:bmax0,1:lmax_win)) !Defined over ACTPol's full ell range
   do i=1,bmax0
      read(lun,*) (win_func_te(i,il), il=2,lmax_win)
   enddo
   close(lun)

   ! EE
   call get_free_lun(lun)
   open(unit=lun,file=bblee_filename,form='formatted',status='unknown',action='read')
    allocate(win_func_ee(1:bmax0,1:lmax_win)) !Defined over ACTPol's full ell range
    do i=1,bmax0
       read(lun,*) (win_func_ee(i,il), il=2,lmax_win)
    enddo
   close(lun)

   !------------------------------------------------- 
   !Read covariance matrix 
   !-------------------------------------------------
   allocate(covmat(1:nbint, 1:nbint))
   call get_free_lun(lun)
   open(unit=lun,file=cov_filename,form='formatted',status='old',action='read')
   do i=1,nbint
      read(lun,*) (covmat(i,il), il=1,nbint)
   enddo 
   close(lun)

   !-------------------------------------------------
   !Read leakage correction if doing a data run
   !-------------------------------------------------
   if(data_run) then
     call get_free_lun(lun)
     open(unit=lun,file=leakage_correction_filename,form='formatted',status='unknown',action='read')
     allocate(l98(1:nbins_leak),l150(1:nbins_leak),ell(1:nbins_leak))
     do i=1,nbins_leak
        read(lun,*) ell(i), l98(i), l150(i)
     enddo
     close(lun)
   end if

   initialise_actpol_deep = .false.

  END SUBROUTINE actpol_deep_likelihood_init
! ===============================================================================================


! ===============================================================================================
  SUBROUTINE get_fg(a_tsz,a_ksz,xi,a_d,a_s,a_c,a_gd,a_tps,a_ps,a_gted,a_geed,a_ste,a_see,beta_c,fgtt,fgttpol,fgpol, lmax_win)
! -----------------------------------------------------------------------------------------------
  real(8), intent(in) :: a_tsz,a_ksz,xi,a_d,a_s,a_c,a_gd,a_tps,a_ps,a_gted,a_geed,a_ste,a_see,beta_c
  integer, intent(in) :: lmax_win
  real(8), intent(inout) :: fgtt(nspecf,2:lmax_win),fgttpol(nspecf,2:lmax_win),fgpol(nspecf,2:lmax_win)
  real(8) :: beta_d, alpha_s, beta_g, alpha_gs
  real(8) :: feff,f0,freq(nnu),fszcorr(nnu)
  real(8) :: fdust(nnu),fsyn(nnu),fsz(nnu)
  real(8) :: sz_corr, planckfunctionratio_corr,flux2tempratio_corr
  real(8) :: planckratiod(nnu),planckratiodg(nnu),fluxtempd(nnu),fluxtemps(nnu)
  integer :: lun,il,i,j,k
  character(len=30)  :: Format

  feff     = 150.d0

  fsz(1)   = 98.4d0
  fsz(2)   = 150.1d0
  fdust(1) = 98.8d0
  fdust(2) = 151.2d0
  fsyn(1)  = 95.8d0
  fsyn(2)  = 147.2d0

  ! Set spectral indexes
  beta_g = 1.5d0
  beta_d = beta_c 
  alpha_s = -0.5d0
  alpha_gs = -1.d0

  do i=1,nnu
       call sz_func(fsz(i),sz_corr)
       fszcorr(i) = sz_corr
  end do
  call sz_func(feff,sz_corr)
  f0 = sz_corr

  do i=1,nnu
       call planckfunctionratio(fdust(i),feff,planckfunctionratio_corr)
       planckratiod(i) = planckfunctionratio_corr

       call planckfunctionratiog(fdust(i),feff,planckfunctionratio_corr)
       planckratiodg(i) = planckfunctionratio_corr
 
       call flux2tempratio(fdust(i), feff, flux2tempratio_corr)
       fluxtempd(i) = flux2tempratio_corr
  
       call flux2tempratio(fsyn(i),feff,flux2tempratio_corr)
       fluxtemps(i) = flux2tempratio_corr
  end do


! Calculate foregrounds
  fgtt(nspecf,2:lmax_win)=0.d0
  fgpol(nspecf,2:lmax_win)=0.d0
  fgttpol(nspecf,2:lmax_win)=0.d0

  do i=1,nspecf
     do il=2,lmax_win 
        if(i==1) then ! 95x95
           fgtt(i,il)  =  a_c*clc(il)*(fdust(1)/feff)**(2.d0*beta_c)*(planckratiod(1)*fluxtempd(1))**2.d0 & !clustered-cib
                          +a_d*clp(il)*(fdust(1)/feff)**(2.d0*beta_d)*(planckratiod(1)*fluxtempd(1))**2.d0 & !poisson-cib
                          +a_s*clp(il)*(fsyn(1)/feff)**(2.d0*alpha_s)*fluxtemps(1)**2.d0 & !poisson-radio 
                          +a_gd*clgt(il)*(fdust(1)/feff)**(2.d0*beta_g)*(planckratiodg(1)*fluxtempd(1))**2.d0 &!galactic dust
                          +a_tsz*cltsz(il)*fszcorr(1)*fszcorr(1)/(f0*f0)& !tsz
                          +a_ksz*clksz(il) & !ksz
                          -2.d0*sqrt(a_c*a_tsz)*xi*clszcib(il)*((fdust(1)**beta_c*fszcorr(1)*planckratiod(1)*fluxtempd(1))/(feff**beta_c*f0)) !tszxcib

           fgpol(i,il)   =  a_ps*clp(il)*(fsyn(1)/feff)**(2.d0*alpha_s)*fluxtemps(1)**2.d0 & !polarized radio poisson 
                           +a_geed*clgp(il)*(fdust(1)/feff)**(2.d0*beta_g)*(planckratiodg(1)*fluxtempd(1))**2.d0 &!galactic dust
                           +a_see*clsp(il)*(fsyn(1)/feff)**(2.d0*alpha_gs)*fluxtemps(1)**2.d0 !synchrotron
  
           fgttpol(i,il) =  a_tps*clp(il)*(fsyn(1)/feff)**(2.d0*alpha_s)*fluxtemps(1)**2.d0 & !temperature/polarized radio poisson
                           +a_gted*clgp(il)*(fdust(1)/feff)**(2.d0*beta_g)*(planckratiodg(1)*fluxtempd(1))**2.d0 & !galactic dust
                           +a_ste*clsp(il)*(fsyn(1)/feff)**(2.d0*alpha_gs)*fluxtemps(1)**2.d0 !synchrotron
 
        else if(i==2) then ! 95x150
           fgtt(i,il)  =  a_c*clc(il)*((fdust(1)*fdust(2)/feff**2.d0)**beta_c)*planckratiod(1)*planckratiod(2)*fluxtempd(1)*fluxtempd(2) & !clustered-cib
                          +a_d*clp(il)*((fdust(1)*fdust(2)/feff**2.d0)**beta_d)*planckratiod(1)*planckratiod(2)*fluxtempd(1)*fluxtempd(2) & !poisson-cib
                          +a_s*clp(il)*((fsyn(1)*fsyn(2)/feff**2.d0)**alpha_s)*fluxtemps(1)*fluxtemps(2) & !poisson-radio 
                          +a_gd*clgt(il)*((fdust(1)*fdust(2)/feff**2.d0)**beta_g)*planckratiodg(1)*planckratiodg(2)*fluxtempd(1)*fluxtempd(2) &!galactic dust
                          +a_tsz*cltsz(il)*fszcorr(1)*fszcorr(2)/(f0*f0)& !tsz
                          +a_ksz*clksz(il) & !ksz
                          -2.d0*sqrt(a_c*a_tsz)*xi*clszcib(il)*((fdust(1)**beta_c*fszcorr(2)*planckratiod(1)*fluxtempd(1)+fdust(2)**beta_c*fszcorr(1)*planckratiod(2)*fluxtempd(2))/(2*feff**beta_c*f0)) !tszxcib

           fgpol(i,il)   = a_ps*clp(il)*((fsyn(1)*fsyn(2)/feff**2.d0)**alpha_s)*fluxtemps(1)*fluxtemps(2) &!polarized radio poisson 
                           +a_geed*clgp(il)*((fdust(1)*fdust(2)/feff**2.d0)**beta_g)*planckratiodg(1)*planckratiodg(2)*fluxtempd(1)*fluxtempd(2) &!galactic dust
                           +a_see*clsp(il)*((fsyn(1)*fsyn(2)/feff**2.d0)**alpha_gs)*fluxtemps(1)*fluxtemps(2) !synchrotron
 
           fgttpol(i,il) = a_tps*clp(il)*((fsyn(1)*fsyn(2)/feff**2.d0)**alpha_s)*fluxtemps(1)*fluxtemps(2) &!temperature/polarized radio poisson
                          +a_gted*clgp(il)*((fdust(1)*fdust(2)/feff**2.d0)**beta_g)*planckratiodg(1)*planckratiodg(2)*fluxtempd(1)*fluxtempd(2) &!galactic dust
                           +a_ste*clsp(il)*((fsyn(1)*fsyn(2)/feff**2.d0)**alpha_gs)*fluxtemps(1)*fluxtemps(2) !synchrotron
 
       else if(i==3) then ! 150x150
           fgtt(i,il)  = a_c*clc(il)*(fdust(2)/feff)**(2.d0*beta_c)*(planckratiod(2)*fluxtempd(2))**2.d0 & !clustered-cib
                          +a_d*clp(il)*(fdust(2)/feff)**(2.d0*beta_d)*(planckratiod(2)*fluxtempd(2))**2.d0 & !poisson-cib
                          +a_s*clp(il)*(fsyn(2)/feff)**(2.d0*alpha_s)*fluxtemps(2)**2.d0 & !poisson-radio 
                          +a_gd*clgt(il)*(fdust(2)/feff)**(2.d0*beta_g)*(planckratiodg(2)*fluxtempd(2))**2.d0 &!galactic dust
                          +a_tsz*cltsz(il)*fszcorr(2)*fszcorr(2)/(f0*f0)& !tsz
                          +a_ksz*clksz(il) & !ksz 
                          -2.d0*sqrt(a_c*a_tsz)*xi*clszcib(il)*((fdust(2)**beta_c*fszcorr(2)*planckratiod(2)*fluxtempd(2))/(feff**beta_c*f0)) !tszxcib
  
           fgpol(i,il)   = a_ps*clp(il)*(fsyn(2)/feff)**(2.d0*alpha_s)*fluxtemps(2)**2.d0 &!polarized radio poisson 
                           +a_geed*clgp(il)*(fdust(2)/feff)**(2.d0*beta_g)*(planckratiodg(2)*fluxtempd(2))**2.d0 &!galactic dust
                           +a_see*clsp(il)*(fsyn(2)/feff)**(2.d0*alpha_gs)*fluxtemps(2)**2.d0 !synchrotron
 
           fgttpol(i,il) = a_tps*clp(il)*(fsyn(2)/feff)**(2.d0*alpha_s)*fluxtemps(2)**2.d0 &!temperature/polarized radio poisson
                           +a_gted*clgp(il)*(fdust(2)/feff)**(2.d0*beta_g)*(planckratiodg(2)*fluxtempd(2))**2.d0 &!galactic dust
                           +a_ste*clsp(il)*(fsyn(2)/feff)**(2.d0*alpha_gs)*fluxtemps(2)**2.d0 !synchrotron
 
      end if
     enddo
  enddo  


    END SUBROUTINE get_fg
! ===============================================================================================

! =============================================================================================================================================================
  SUBROUTINE actpol_deep_likelihood_compute(cl, a_tsz, a_ksz, xi, a_d, a_s, a_c, a_gd, a_tps, a_ps, a_gted, a_geed, a_ste, a_see, beta_c, ct1, ct2, yp1, yp2, tt_lmax, nbint, nbintt, nbinte, nbinee, lmax_win, bmax0, b1tt, b1te, b1ee, use_cls, use_tt, use_te, use_ee, save_test_output, test_save_dir, test_save_root, invcov, data_run, like)
! -------------------------------------------------------------------------------------------------------------------------------------------------------------

  implicit none
  integer, intent(in) :: tt_lmax, nbint, lmax_win, bmax0, b1tt, b1te, b1ee
  integer :: nbintt, nbinte, nbinee
  real(8), intent(in) :: cl(2:tt_lmax,1:3)
  real(8) :: cl_tt(nbintt),cl_te(nbinte),cl_ee(nbinee)
  real(8), intent(in) :: a_tsz,a_ksz,xi,a_d,a_s,a_c,a_gd,a_tps,a_ps,a_gted,a_geed,a_ste,a_see,beta_c,ct1,ct2,yp1,yp2
  real(8), intent(inout) :: like
  integer :: bin_no,lun,il,i,j,info,k, bmin, bmax
  integer :: nbin_diff ! difference in number of bins between TT and TE
  integer :: b1  ! first bin used for TE and EE - needed for leakage correction
  real(8) :: cltt(2:lmax_win),clte(2:lmax_win),clee(2:lmax_win),X_theory(nspectot,2:lmax_win)
  real(8) :: diffs(nbint),X_model(nbint) 
  real(8) :: theory(nbinte*nspectot) ! tmp to hold binned theory, with all spectra using # bins for TE - used for leakage correction
  real(8) :: dlnlike
  real(8) :: a1, a2 ! for leakage; leakage errors in covmat, so these are set to one
  real(8), dimension(:), allocatable :: tmp, tmp_model ! used for intermediate steps
  real(8) :: fgtt(nspecf,2:lmax_win),fgttpol(nspecf,2:lmax_win), fgpol(nspecf,2:lmax_win)
  logical :: use_cls, use_tt, use_te, use_ee, save_test_output, invcov, data_run
  character(len=240) :: test_save_dir, test_save_root

  a1 = 1.d0
  a2 = 1.d0

! calculate FG
  call get_fg(a_tsz,a_ksz,xi,a_d,a_s,a_c,a_gd,a_tps,a_ps,a_gted,a_geed,a_ste,a_see,beta_c,fgtt,fgttpol,fgpol, lmax_win)


! save theory
  if(save_test_output) then

    call get_free_lun(lun)

    open(unit=lun, file=trim(test_save_dir)//trim(test_save_root)//'_CMB.txt', status='replace', action='write')
    do i = 2, lmax_win
      write(lun, *) cl(i,1), cl(i,2), cl(i,3)
    enddo
    close(lun)

    call get_free_lun(lun)
    open(unit=lun, file=trim(test_save_dir)//trim(test_save_root)//'_fgTT.txt', status='replace', action='write')
    do i = 2, lmax_win
      write(lun, *) fgtt(1, i), fgtt(2, i), fgtt(3, i)
    enddo
    close(lun)

    call get_free_lun(lun)
    open(unit=lun, file=trim(test_save_dir)//trim(test_save_root)//'_fgTE.txt', status='replace', action='write')
    do i = 2, lmax_win
      write(lun, *) fgttpol(1, i), fgttpol(2, i), fgttpol(3, i)
    enddo
    close(lun)

    call get_free_lun(lun)
    open(unit=lun, file=trim(test_save_dir)//trim(test_save_root)//'_fgEE.txt', status='replace', action='write')
    do i = 2, lmax_win
      write(lun, *) fgpol(1, i), fgpol(2, i), fgpol(3, i)
    enddo
    close(lun)

  end if


! Get TT, TE, EE theory (actually in Dl's by default) 
  cltt(2:lmax_win)=0.d0 
  clte(2:lmax_win)=0.d0
  clee(2:lmax_win)=0.d0
  cltt(2:tt_lmax)=cl(2:tt_lmax,1)
  clte(2:tt_lmax)=cl(2:tt_lmax,2)
  clee(2:tt_lmax)=cl(2:tt_lmax,3)

!-----------------------------------------------
!Calculate CMB+Fg 
!-----------------------------------------------
  X_theory= 0.d0

  X_theory(1,2:lmax_win) = (cltt(2:lmax_win)+fgtt(1,2:lmax_win))    ! TT 90x90
  X_theory(2,2:lmax_win) = (cltt(2:lmax_win)+fgtt(2,2:lmax_win))    ! TT 90x150
  X_theory(3,2:lmax_win) = (cltt(2:lmax_win)+fgtt(3,2:lmax_win))    ! TT 150x150
  X_theory(4,2:lmax_win) = (clte(2:lmax_win)+fgttpol(1,2:lmax_win)) ! TE 90x90
  X_theory(5,2:lmax_win) = (clte(2:lmax_win)+fgttpol(2,2:lmax_win)) ! TE 90x150
  X_theory(6,2:lmax_win) = (clte(2:lmax_win)+fgttpol(2,2:lmax_win)) ! TE 150x90
  X_theory(7,2:lmax_win) = (clte(2:lmax_win)+fgttpol(3,2:lmax_win)) ! TE 150x150
  X_theory(8,2:lmax_win) = (clee(2:lmax_win)+fgpol(1,2:lmax_win))   ! EE 90x90
  X_theory(9,2:lmax_win) = (clee(2:lmax_win)+fgpol(2,2:lmax_win))   ! EE 90x150
  X_theory(10,2:lmax_win)= (clee(2:lmax_win)+fgpol(3,2:lmax_win))   ! EE 150x150

!Get theory in Cls if necessary
  if(use_cls) then
    do i=2,lmax_win
      X_theory(1:nspectot,i)=X_theory(1:nspectot,i)/real(i*(i+1.)/2./pi)
    end do
  end if

!-----------------------------------------------
! Get binned model
!-----------------------------------------------
  X_model= 0.d0

  ! if we only want to use the binned model from b1:bmax0, need tmp array to hold full result of binning
  allocate(tmp_model(1:3*bmax0))
  tmp_model = 0.d0

  ! to properly apply leakage correction, we need to keep the same number of bins in TT as we have in TE and EE
  ! so theory will hold the binned model, with nbinte bins for each spectrum (including TT)
  ! here, set nbintt = nbinte; then later, change it back by setting nbintt = nbintt - nbin_diff

  nbin_diff = b1tt - b1te
  nbintt = nbinte
  b1 = b1te

  ! TT 90
  tmp_model(1:bmax0) = MATMUL(win_func_tt(1:bmax0,2:lmax_win), X_theory(1,2:lmax_win))
  theory(1:nbintt) = tmp_model(b1:bmax0)

  ! TT 90x150
  tmp_model(1:bmax0) = MATMUL(win_func_tt(1:bmax0,2:lmax_win), X_theory(2,2:lmax_win))
  theory(nbintt+1:2*nbintt) = tmp_model(b1:bmax0)

  ! TT 150
  tmp_model(1:bmax0) = MATMUL(win_func_tt(1:bmax0,2:lmax_win), X_theory(3,2:lmax_win))
  theory(2*nbintt+1:3*nbintt) = tmp_model(b1:bmax0)

  ! TE 90
  tmp_model(1:bmax0) = MATMUL(win_func_te(1:bmax0,2:lmax_win), X_theory(4,2:lmax_win))
  theory(3*nbintt+1:3*nbintt+nbinte) = tmp_model(b1:bmax0)

  ! TE 90x150
  tmp_model(1:bmax0) = MATMUL(win_func_te(1:bmax0,2:lmax_win), X_theory(5,2:lmax_win))
  theory(nbinte+3*nbintt+1:3*nbintt+2*nbinte) = tmp_model(b1:bmax0)

  ! TE 150x90
  tmp_model(1:bmax0) = MATMUL(win_func_te(1:bmax0,2:lmax_win), X_theory(6,2:lmax_win))
  theory(2*nbinte+3*nbintt+1:3*nbintt+3*nbinte) = tmp_model(b1:bmax0)

  ! TE 150
  tmp_model(1:bmax0) = MATMUL(win_func_te(1:bmax0,2:lmax_win), X_theory(7,2:lmax_win))
  theory(3*nbinte+3*nbintt+1:3*nbintt+4*nbinte) = tmp_model(b1:bmax0)

  ! EE 90
  tmp_model(1:bmax0) = MATMUL(win_func_ee(1:bmax0,2:lmax_win), X_theory(8,2:lmax_win))
  theory(4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+nbinee) = tmp_model(b1:bmax0)

  ! EE 90x150
  tmp_model(1:bmax0) = MATMUL(win_func_ee(1:bmax0,2:lmax_win), X_theory(9,2:lmax_win))
  theory(nbinee+4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+2*nbinee) = tmp_model(b1:bmax0)

  ! EE 150
  tmp_model(1:bmax0) = MATMUL(win_func_ee(1:bmax0,2:lmax_win), X_theory(10,2:lmax_win))
  theory(2*nbinee+3*nbintt+4*nbinte+1:3*nbintt+4*nbinte+3*nbinee) = tmp_model(b1:bmax0)


!-----------------------------------------------
! if doing a data run, correct theory for leakage
!-----------------------------------------------
  if (data_run) then
    ! te 90x90 = (te 90x90) + (tt 90x90)*l98
    theory(3*nbintt+1:3*nbintt+nbinte) =  theory(3*nbintt+1:3*nbintt+nbinte)+theory(1:nbintt)*a1*l98(1:nbinte)

    ! te 90x150 = (te 90x150) + (tt 90x150)*l150
    theory(nbinte+3*nbintt+1:3*nbintt+2*nbinte) = theory(nbinte+3*nbintt+1:3*nbintt+2*nbinte)+theory(nbintt+1:2*nbintt)*a2*l150(1:nbinte)

    ! te 150x90 = (te 150x90) + (tt 90x150)*l98
    theory(2*nbinte+3*nbintt+1:3*nbintt+3*nbinte) = theory(2*nbinte+3*nbintt+1:3*nbintt+3*nbinte)+theory(nbintt+1:2*nbintt)*a1*l98(1:nbinte)

    ! te 150x150 = (te 150x150) + (tt 150x150)*l150
    theory(3*nbinte+3*nbintt+1:3*nbintt+4*nbinte) = theory(3*nbinte+3*nbintt+1:3*nbintt+4*nbinte)+theory(2*nbintt+1:3*nbintt)*a2*l150(1:nbinte)

    ! ee 90x90 = (ee 90x90) + 2*(te 90x90)*l98 + (tt 90x90)*(l98**2)
    theory(4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+nbinee) = theory(4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+nbinee)+2*theory(3*nbintt+1:3*nbintt+nbinte)*a1*l98(1:nbinte)+theory(1:nbintt)*(a1*l98(1:nbinte))**2.d0

    ! ee 90x150 = (ee 90x150) + (te 90x150)*l98 + (te 150x90)*l150 + (tt 90x150)*l98*l150
    theory(nbinee+4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+2*nbinee) = theory(nbinee+4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+2*nbinee)+theory(nbinte+3*nbintt+1:3*nbintt+2*nbinte)*a1*l98(1:nbinte)+theory(2*nbinte+3*nbintt+1:3*nbintt+3*nbinte)*a2*l150(1:nbinte)+theory(nbintt+1:2*nbintt)*a1*l98(1:nbinte)*a2*l150(1:nbinte)

    ! ee 150x150 = (ee 150x150) + 2*(te 150x150)*l150 + (tt 150x150)*(l150**2)
    theory(2*nbinee+3*nbintt+4*nbinte+1:3*nbintt+4*nbinte+3*nbinee) = theory(2*nbinee+3*nbintt+4*nbinte+1:3*nbintt+4*nbinte+3*nbinee)+2*theory(3*nbinte+3*nbintt+1:3*nbintt+4*nbinte)*a2*l150(1:nbinte)+theory(2*nbintt+1:3*nbintt)*(a2*l150(1:nbinte))**2.d0

  end if


!-----------------------------------------------
! trim the TT spectra to use only nbintt bins (remove first 2 bins from TT spectra), and put binned theory into X_model
!-----------------------------------------------
  nbintt = nbintt - nbin_diff

  ! TT 90x90
  bmin = nbin_diff ! nbin_diff should be 2
  bmax = bmin + nbintt
  X_model(1:nbintt) = theory(1+bmin:bmax)
  ! TT 90x150
  bmin = bmax + nbin_diff
  bmax = bmin + nbintt
  X_model(nbintt+1:2*nbintt) = theory(1+bmin:bmax)
  ! TT 150x150
  bmin = bmax + nbin_diff
  bmax = bmin + nbintt
  X_model(2*nbintt+1:3*nbintt) = theory(1+bmin:bmax)
  ! TE 90x90
  bmin = bmax 
  bmax = bmin + nbinte
  X_model(3*nbintt+1:3*nbintt+nbinte) = theory(1+bmin:bmax)
  ! TE 90x150
  bmin = bmax 
  bmax = bmin + nbinte
  X_model(nbinte+3*nbintt+1:3*nbintt+2*nbinte) = theory(1+bmin:bmax)
  ! TE 150x90
  bmin = bmax 
  bmax = bmin + nbinte
  X_model(2*nbinte+3*nbintt+1:3*nbintt+3*nbinte) = theory(1+bmin:bmax)
  ! TE 150x150
  bmin = bmax 
  bmax = bmin + nbinte
  X_model(3*nbinte+3*nbintt+1:3*nbintt+4*nbinte) = theory(1+bmin:bmax)
  ! EE 90x90
  bmin = bmax 
  bmax = bmin + nbinee
  X_model(4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+nbinee) = theory(1+bmin:bmax)
  ! EE 90x150
  bmin = bmax 
  bmax = bmin + nbinee
  X_model(nbinee+4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+2*nbinee) = theory(1+bmin:bmax)
  ! EE 150x150 
  bmin = bmax 
  bmax = bmin + nbinee
  X_model(2*nbinee+3*nbintt+4*nbinte+1:3*nbintt+4*nbinte+3*nbinee) = theory(1+bmin:bmax)


!-----------------------------------------------
! Calibrate
!-----------------------------------------------
  X_model(1:nbintt)=X_model(1:nbintt)*ct1*ct1
  X_model(nbintt+1:2*nbintt)=X_model(nbintt+1:2*nbintt)*ct1*ct2
  X_model(2*nbintt+1:3*nbintt)=X_model(2*nbintt+1:3*nbintt)*ct2*ct2
  X_model(3*nbintt+1:3*nbintt+nbinte)=X_model(3*nbintt+1:3*nbintt+nbinte)*ct1*ct1*yp1
  X_model(nbinte+3*nbintt+1:3*nbintt+2*nbinte)=X_model(nbinte+3*nbintt+1:3*nbintt+2*nbinte)*ct1*ct2*yp2
  X_model(2*nbinte+3*nbintt+1:3*nbintt+3*nbinte)=X_model(2*nbinte+3*nbintt+1:3*nbintt+3*nbinte)*ct1*ct2*yp1
  X_model(3*nbinte+3*nbintt+1:3*nbintt+4*nbinte)=X_model(3*nbinte+3*nbintt+1:3*nbintt+4*nbinte)*ct1*ct2*yp2
  X_model(4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+nbinee)=X_model(4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+nbinee)*ct1*ct1*yp1*yp1
  X_model(nbinee+4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+2*nbinee)=X_model(nbinee+4*nbinte+3*nbintt+1:3*nbintt+4*nbinte+2*nbinee)*ct1*ct2*yp1*yp2
  X_model(2*nbinee+3*nbintt+4*nbinte+1:3*nbintt+4*nbinte+3*nbinee)=X_model(2*nbinee+3*nbintt+4*nbinte+1:3*nbintt+4*nbinte+3*nbinee)*ct2*ct2*yp2*yp2



!-----------------------------------------------
!Basic chisq 
!-----------------------------------------------
   like = 0.d0
   diffs = b_dat - X_model 

  !Select data
  !Only TT
  if((use_tt .eqv. .true.) .and. (use_te .eqv. .false.) .and. (use_ee .eqv. .false.)) then
       bin_no=nbintt*nspectt
       allocate(fisher(bin_no,bin_no),diff_vec(bin_no),tmp(bin_no))
       diff_vec(:)=diffs(1:bin_no)
       fisher(:,:)=covmat(1:bin_no,1:bin_no)
  !Only TE
  else if((use_tt .eqv. .false.) .and. (use_te .eqv. .true.) .and. (use_ee .eqv. .false.)) then
       bin_no=nbinte*nspecte
       allocate(fisher(bin_no,bin_no),diff_vec(bin_no),tmp(bin_no))
       diff_vec(:)=diffs(nbintt*nspectt+1:nbintt*nspectt+bin_no)
       fisher(:,:)=covmat(nbintt*nspectt+1:nbintt*nspectt+bin_no,nbintt*nspectt+1:nbintt*nspectt+bin_no)
  !Only EE
  else if((use_tt .eqv. .false.) .and. (use_te .eqv. .false.) .and. (use_ee .eqv. .true.)) then
       bin_no=nbinee*nspecee
       allocate(fisher(bin_no,bin_no),diff_vec(bin_no),tmp(bin_no))
       diff_vec(:)=diffs(nbintt*nspectt+nbinte*nspecte+1:nbintt*nspectt+nbinte*nspecte+bin_no)
       fisher(:,:)=covmat(nbintt*nspectt+nbinte*nspecte+1:nbintt*nspectt+nbinte*nspecte+bin_no,nbintt*nspectt+nbinte*nspecte+1:nbintt*nspectt+nbinte*nspecte+bin_no)
  !All 
  else if ((use_tt .eqv. .true.) .and. (use_te .eqv. .true.) .and. (use_ee .eqv. .true.)) then
       bin_no=nbint
       allocate(fisher(bin_no,bin_no),diff_vec(bin_no),tmp(bin_no))
       diff_vec(:)=diffs(:)
       fisher(:,:)=covmat(:,:) !baseline
  else 
     write(*,*) 'Fail: no possible options chosen'
  endif

!Invert covmat if necessary
  if(invcov .eqv. .false.) then
     call dpotrf('U',bin_no,fisher,bin_no,info)
     if(info.ne.0)then
        print*, ' info in dpotrf =', info
        stop
     endif
     call dpotri('U',bin_no,fisher,bin_no,info)
     if(info.ne.0)then
        print*, ' info in dpotri =', info
        stop
     endif
     do i=1,bin_no
        do j=i,bin_no
           fisher(j,i)=fisher(i,j)
        enddo
     enddo
  endif
 
  tmp=matmul(fisher,diff_vec)
  dlnlike=sum(tmp*diff_vec)
  deallocate(fisher,diff_vec,tmp, tmp_model)
    
  like = dlnlike/2.d0
    
  end SUBROUTINE actpol_deep_likelihood_compute
! ===============================================================================================
 
END MODULE ACTPol_deep
! ===============================================================================================
