!===========================================================
MODULE ACTPol_like

  use Options
  use ACTPol_deep
  use Fg_subroutines
  use Foregrounds_loading

  implicit none
contains

 ! =======================================================================
  SUBROUTINE ACTPol_likelihood_init(spec_fname, cov_fname, bbltt_fname, bblte_fname, bblee_fname, data_dir, act_data_dir, nbint, lmax_win, bmax0, b1tt, b1te, b1ee, data_run, leakage_correction_fname)

    character(len=*) :: spec_fname, cov_fname
    character(len=*) :: bbltt_fname
    character(len=*) :: bblte_fname
    character(len=*) :: bblee_fname
    character(len=*) :: data_dir, act_data_dir
    character(len=*) :: leakage_correction_fname
    integer :: nbint, lmax_win, bmax0, b1tt, b1te, b1ee
    logical :: data_run
     
    print*, 'Initializing ACTPol likelihood'
    call fg_init(data_dir, lmax_win)
    call actpol_deep_likelihood_init(spec_fname, cov_fname, bbltt_fname, bblte_fname, bblee_fname, data_dir, act_data_dir, nbint, lmax_win, bmax0, b1tt, b1te, b1ee, data_run, leakage_correction_fname)

  END SUBROUTINE ACTPol_likelihood_init
  ! ======================================================================


  ! ==================================================================================================================================
  SUBROUTINE actpol_likelihood_compute(cell, a_tsz, a_ksz, xi, a_d, a_s, a_c, a_gd, a_tps, a_ps, a_gted, a_geed, a_ste, a_see, beta_c, ct1, ct2, yp1, yp2, tt_lmax, nbint, nbintt, nbinte, nbinee, lmax_win, bmax0, b1tt, b1te, b1ee, use_cls, use_tt, use_te, use_ee, save_test_output, test_save_dir, test_save_root, invcov, data_run, like)
  !-----------------------------------------------------------------------------------------------------------------------------------
    implicit none
    integer, intent(in) :: tt_lmax, nbint, nbintt, nbinte, nbinee, lmax_win, bmax0, b1tt, b1te, b1ee
    real(8), intent(in) :: cell(2:tt_lmax,1:3) 
    real(8), intent(in) :: a_tsz, a_ksz, xi, a_d, a_s, a_c, a_gd, a_tps, a_ps, a_gted, a_geed, a_ste, a_see, beta_c, ct1, ct2, yp1, yp2
    logical, intent(in) :: use_cls, use_tt, use_te, use_ee, save_test_output, invcov, data_run
    character(len=240), intent(in) :: test_save_dir, test_save_root
    real(8)  :: liked
    real(8), intent(inout) :: like

    liked = 0.d0
    like  = 0.d0

!$omp parallel sections DEFAULT(SHARED)
!$omp section
    if (use_deep .eq. .true.) then
        call actpol_deep_likelihood_compute(cell, a_tsz, a_ksz, xi, a_d, a_s, a_c, a_gd, a_tps, a_ps, a_gted, a_geed, a_ste, a_see, beta_c, ct1, ct2, yp1, yp2, tt_lmax, nbint, nbintt, nbinte, nbinee, lmax_win, bmax0, b1tt, b1te, b1ee, use_cls, use_tt, use_te, use_ee, save_test_output, test_save_dir, test_save_root, invcov, data_run, liked)
    end if
!$omp end parallel sections

    like= liked

   print *, "----------------------------------------"
   print *, 'ACTPol deep chi2 =', 2.d0*liked

  END SUBROUTINE actpol_likelihood_compute
  ! =====================================================================================================================================

END MODULE ACTPol_like

