! =======================================================================================
MODULE foregrounds_loading
! =======================================================================================
  use Options
  use Fg_subroutines

  implicit none
  logical :: initialise_foregrounds=.true.
  real(8), dimension (:), allocatable :: cltsz,clksz,clszcib,clgt,clp,clc,clgp,clsp
  public :: fg_init

   contains

   ! ===================================================================================
   SUBROUTINE fg_init(data_dir, lmax_win)
   ! ===================================================================================

    implicit none
    character(len=240) :: data_dir
    integer :: lmax_win
    integer  :: lun,il
    real(8)  :: dummy
    allocate(cltsz(2:lmax_win),clksz(2:lmax_win),clszcib(2:lmax_win))
    allocate(clp(2:lmax_win),clgt(2:lmax_win),clc(2:lmax_win),clgp(2:lmax_win),clsp(2:lmax_win))

    !Load tSZ template
    call get_free_lun(lun)
    open(unit=lun,file=trim(data_dir)//'Fg/cl_tsz_150_bat.dat',form='formatted',status='unknown')
    do il=2,lmax_win
       read(lun,*) dummy,cltsz(il)
       cltsz(il) = cltsz(il)/5.59550d0 !This normalizes the tSZ spectrum to 1 at l=3000
    enddo
    close(lun)

    !Load kSZ template
    call get_free_lun(lun)
    open(unit=lun,file=trim(data_dir)//'Fg/cl_ksz_bat.dat',form='formatted',status='unknown')
    do il=2,lmax_win
       read(lun,*) dummy,clksz(il)
       clksz(il) = clksz(il)/1.51013d0 !This normalizes the kSZ spectrum to 1 at l=3000
    enddo
    close(lun)

    !Define Poisson spectrum 
    do il=2,lmax_win
       clp(il)=(il*(il+1.d0)/3000.d0/3001.d0)
    enddo

    !Define Clustered cib
    do il=2,lmax_win
       clc(il)=(il*(il+1.d0)/3000.d0/3001.d0)*(il/3000.d0)**(-1.2d0)
    enddo

    !Load 217 CIB template extrapolated for ACT/SPT
    call get_free_lun(lun)
    clc(2:lmax_win) = 0.d0
    open(unit=lun,file=trim(data_dir)//'Fg/cib_extra.dat',form='formatted',status='unknown')
    do il=2,lmax_win
       read(lun,*) dummy,clc(il)
    enddo
    close(lun)

    !Load tSZ-CIB template
    call get_free_lun(lun)
    clszcib(2:lmax_win) = 0.d0
    open(unit=lun,file=trim(data_dir)//'Fg/sz_x_cib_template.dat',form='formatted',status='unknown')
    do il=2,lmax_win-1
       read(lun,*) dummy,clszcib(il)
    enddo
    close(lun)

    !Define Galactic dust template in TT
    do il=2,lmax_win
       clgt(il)=(il/500.d0)**(-0.6d0)
    enddo

    !Define Galactic dust in TE/EE
    do il=2,lmax_win
       clgp(il)=(il/500.d0)**(-0.4d0)
    enddo

    !Define synchrotron in TE/EE
    do il=2,lmax_win
       clsp(il)=(il/500.d0)**(-0.7d0)
    enddo

    initialise_foregrounds=.false.

    END SUBROUTINE fg_init 
    !=====================================================================================

END MODULE foregrounds_loading
!=========================================================================================
