import sys
import camb.correlations
import numpy as np

# get delensed CAMB theory curve from input lensed and unlensed theory

def get_delensed(tt_unlen, te_unlen, ee_unlen, ldd_theo, cldd_theo, L, N_L, klmin, klmax, load_dls=True, unit='uk', lmax_ret=8000):
	kappa_lmin = int(klmin)
	kappa_lmax = int(klmax)
	
	# want theory to start at l = 0
	ldd_th = np.zeros(len(ldd_theo)+int(ldd_theo[0]))
	ldd_th[0] = 0.
	ldd_th[1] = 1.
	ldd_th[int(ldd_theo[0]):] = np.array(ldd_theo)
	# fill in unlensed theory and lensing potential 
	cldd_th = np.zeros(len(ldd_th))
	tt_th = np.zeros(len(ldd_th))
	te_th = np.zeros(len(ldd_th))
	ee_th = np.zeros(len(ldd_th))
	idx = int(ldd_theo[0])
	cldd_th[idx:] = np.array(cldd_theo)
	tt_th[idx:] = np.array(tt_unlen)
	te_th[idx:] = np.array(te_unlen)
	ee_th[idx:] = np.array(ee_unlen)
	# fill in the noise curve so it also starts at L=0 and is same size as theory
	NL = np.zeros(ldd_th.shape)
	idx = int(L[0])
	NL[:idx] = np.inf
	if len(N_L) >= len(NL) + idx:
		NL[idx:] = np.array(N_L[:len(NL)+idx])
	else:
		NL[idx:len(N_L)+idx] = np.array(N_L)
		NL[len(N_L)+idx:] = np.inf		
	L = ldd_th.copy()   
 
	def _weiner1d_filter(signal, noise):
		return (signal/(signal+noise))*signal


	# get residual lensing potential
	# NL is in the same form as clkk ; want to use cldd form
	#NL *= (2. * np.pi) / 4.

	clkk = cldd_th * (2 * np.pi / 4)
	clkk_filt      = _weiner1d_filter(clkk, NL)
	cldd_filt = clkk_filt * 4 / (2*np.pi)
	loc            = np.where(L > kappa_lmax)
	if len(loc[0]) > 0:
		if loc[0][0] > len(cldd_filt):
			cldd_filt[loc[0][0]:] = 0.
	loc            = np.where(L < kappa_lmin)
	cldd_filt[loc] = 0.

	cldd_res = cldd_th-cldd_filt


	# create array to hold delensed spectra
	dls = np.zeros((len(ldd_th),4))
	dls[:,0] = tt_th
	dls[:,1] = ee_th
	dls[:,2] = np.zeros(len(tt_th))
	dls[:,3] = te_th

	camb_ret = camb.correlations.lensed_cls(dls, cldd_res, delta_cls=False, sampling_factor=2.)

	ret         = {}
	ret['dltt'] = camb_ret[:,0]
	ret['dlee'] = camb_ret[:,1]
	ret['dlbb'] = camb_ret[:,2]
	ret['dlte'] = camb_ret[:,3]
	
	return ret['dltt'][1:], ret['dlte'][1:], ret['dlee'][1:]



